'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var oscriptAstWalker = require('oscript-ast-walker');
var fs = require('fs');

class Scope {
  constructor (upper, init) {
    this._upper = upper;
    this._vars = new Map();
    if (init) for (const name in init) this.setOwn(name, init[name]);
  }
  get upper () { return this._upper }
  hasOwn (name) {
    return this._vars.has(name)
  }
  has (name) {
    return !!this._findScope(name)
  }
  getOwn (name) {
    return this._vars.get(name)
  }
  get (name) {
    const scope = this._findScope(name);
    return scope && scope.getOwn(name)
  }
  setOwn (name, value) {
    this._vars.set(name, value);
  }
  set (name, value) {
    const scope = this._findScope(name);
    if (scope) {
      scope.setOwn(name, value);
      return true
    }
  }
  _findScope (name) {
    let current = this;
    do {
      if (current.hasOwn(name)) return current
      current = current._upper;
    } while (current)
  }
}

const orders = ['', 'first', 'second', 'third'];
function locate (order) {
  return `found in the ${orders[order]} argument`
}
function checkType (value, expected, order) {
  const actual = typeof value;
  if (actual !== expected) {
    throw new TypeError(`${expected} expected, but ${actual} ${locate(order)}`)
  }
}
function checkTypeOptional (value, expected, order) {
  const actual = typeof value;
  if (actual !== expected && actual !== 'undefined') {
    throw new TypeError(`${expected} or undefined expected, but ${actual} ${locate(order)}`)
  }
}
function checkType2 (value, expected, expected2, order) {
  const actual = typeof value;
  if (actual !== expected && actual !== expected2) {
    throw new TypeError(`${expected} or ${expected2} expected, but ${actual} ${locate(order)}`)
  }
}
function checkAssoc (value, order) {
  const actual = typeof value;
  if (actual !== 'object' || !value || Array.isArray(value)) {
    throw new TypeError(`assoc expected, but ${actual} ${locate(order)}`)
  }
}
function checkList (value, order) {
  if (!Array.isArray(value)) {
    throw new TypeError(`list expected, but ${typeof value} ${locate(order)}`)
  }
}
function checkQuote (value, order) {
  if (value !== '\'' && value !== '"') {
    throw new RangeError(`' or " expected, but ${typeof value} ${locate(order)}`)
  }
}

class Regex {
  match (text) {
    checkType(text, 'string', 1);
    if (!this._expr) throw new Error('pattern not set')
    return this._expr.test(text)
  }
  setpattern (text) {
    checkType(text, 'string', 1);
    this._expr = new RegExp(text);
  }
}

const defaults = {
  assoc () { return {} },
  boolean () { return false },
  integer () { return 0 },
  list () { return [] },
  long () { return 0 },
  real () { return 0 },
  regex () { return new Regex() },
  string () { return '' }
};
function getDefaultValue (type) {
  const getter = defaults[type];
  return getter && getter()
}

const push = Array.prototype.push;
class InterpreterError extends Error {}
class NotImplementedError extends InterpreterError {}
class RuntimeError extends InterpreterError {
  constructor (message, node) {
    super(message);
    this.node = node;
  }
}
function ignore () {}
let warnings;
function setOptions (options = {}) {
  ({ warnings } = options);
}
const visitors = {};
visitors.Program = (node, scope, walk) => walk(node.body, node, scope);
visitors.PackageDeclaration = () => {
  throw new NotImplementedError('package is not implemented')
};
visitors.ObjectName = (node, scope) => {
  let { raw } = node;
  if (!raw) raw = prepareObjectName(node);
  return scope.get(raw)
};
visitors.ScriptSource = (node, scope, walk) => {
  if (!node.statements) prepareScript(node);
  for (const routine of node.routines) walk(routine, node, scope);
  for (const variable of node.variables) walk(variable, node, scope);
  for (const statement of node.statements) walk(statement, node, scope);
};
visitors.DumpSource = () => {
  throw new NotImplementedError('dump is not implemented')
};
visitors.ScriptDeclaration = (node, scope) => {
  if (!node.statements) prepareScript(node);
  scope.setOwn(node.id.value, node);
};
visitors.FunctionDeclaration = (node, scope) => {
  if (!node.statements) prepareFunction(node);
  scope.setOwn(node.id.value, node);
};
visitors.IfStatement = (node, scope, walk) => {
  if (walk(node.test, node, scope)) {
    for (const statement of node.consequent) walk(statement, node, scope);
  } else {
    if (tryElseIfClauses(node.otherClauses, scope, walk)) return
    for (const statement of node.alternate) walk(statement, node, scope);
  }
};
function tryElseIfClauses (otherClauses, scope, walk) {
  for (const otherClause of otherClauses) {
    if (walk(otherClause.test, otherClause, scope)) {
      for (const statement of otherClause.consequent) walk(statement, otherClause, scope);
      return true
    }
  }
}
visitors.SwitchStatement = (node, scope, walk) => {
  const discriminant = walk(node.discriminant, node, scope);
  node.cases.some(switchCase => trySwitchCase(discriminant, switchCase, scope, walk));
};
function trySwitchCase (discriminant, node, scope, walk) {
  if (!node.tests || node.tests.some(test => walk(test, node, scope) === discriminant)) {
    for (const statement of node.consequent) walk(statement, node, scope);
    return true
  }
}
visitors.WhileStatement = (node, scope, walk) => {
  while (walk(node.test, node, scope)) {
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
  }
};
visitors.RepeatStatement = (node, scope, walk) => {
  do {
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
  } while (walk(node.test, node, scope))
};
visitors.ForStatement = (node, scope, walk) => {
  for (node.init && walk(node.init, node, scope);
    !node.test || walk(node.test, node, scope);
    node.update && walk(node.update, node, scope)) {
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
  }
};
visitors.ForEachStatement = (node, scope, walk) => {
  const name = node.left.value;
  for (const item of walk(node.right, node, scope)) {
    scope.set(name, item);
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
  }
};
visitors.StructuredForStatement = (node, scope, walk) => {
  const name = node.variable.value;
  let index = walk(node.start, node, scope);
  const end = walk(node.end, node, scope);
  const step = node.step ? walk(node.step, node, scope) : 1;
  for (scope.set(name, index); index < end;) {
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
    scope.set(name, node.down ? index -= step : index += step);
  }
};
visitors.BreakStatement = (node, scope) => { scope.break = true; };
visitors.ContinueStatement = (node, scope) => { scope.continue = true; };
visitors.BreakIfStatement = (node, scope, walk) => {
  scope.break = walk(node.test, node, scope);
};
visitors.ContinueIfStatement = (node, scope, walk) => {
  scope.continue = walk(node.test, node, scope);
};
visitors.GotoStatement = (node, scope, walk) => {
  throw new NotImplementedError('goto is not implemented')
};
visitors.LabelStatement = visitors.EmptyStatement = ignore;
visitors.ReturnStatement = (node, scope, walk) => {
  if (node.argument) scope.result = walk(node.argument, node, scope);
};
visitors.VariableDeclaration = (node, scope, walk) => {
  for (const declaration of node.declarations) {
    const { init } = declaration;
    scope.setOwn(declaration.id.value,
      init ? walk(init, declaration, scope) : getDefaultValue(node.variableType));
  }
};
visitors.ConditionalExpression = (node, scope, walk) => {
  return walk(node.test, node, scope)
    ? walk(node.consequent, node, scope)
    : walk(node.alternate, node, scope)
};
visitors.BinaryExpression = (node, scope, walk) => {
  const { left, operator, right } = node;
  let target;
  if (isAssignment(operator)) {
    target = getName(left);
    if (!target) throw new RuntimeError('assignment target is not a variable', left)
  } else {
    target = walk(left, node, scope);
  }
  return binaryOperation[operator](target, walk(right, node, scope), scope, left)
};
function isAssignment (symbol) {
  return symbol === '=' ||  symbol === '+=' || symbol === '-=' ||
    symbol === '*=' || symbol === '&=' || symbol === '|=' || symbol === '^='
}
function getName (node) {
  const { type } = node;
  return type === 'Identifier'
    ? node.value
    : type === 'ObjectName' || type === 'LegacyAlias'
      ? node.raw
      : undefined
}
function setVariable (name, value, scope, target) {
  if (!scope.set(name, value)) {
    if (warnings) throw new RuntimeError('variable not declared', target)
    scope.setOwn(name, value);
  }
}
const binaryOperation = {
  '=': (left, right, scope, target) => setVariable(left, right, scope, target),
  '+': (left, right) => left + right,
  '-': (left, right) => left - right,
  '*': (left, right) => left * right,
  '/': (left, right) => left / right,
  '<': (left, right) => left < right,
  '>': (left, right) => left > right,
  '%': (left, right) => left % right,
  '&': (left, right) => left & right,
  '|': (left, right) => left | right,
  '^': (left, right) => left ^ right,
  '==': (left, right) => left === right,
  '!=': (left, right) => left !== right,
  '<>': (left, right) => left !== right,
  '<=': (left, right) => left <= right,
  '>=': (left, right) => left >= right,
  '&&': (left, right) => left && right,
  '||': (left, right) => left || right,
  in: (left, right) => left in right,
  '+=': (left, right, scope, target) => setVariable(left, scope.get(left) + right, scope, target),
  '-=': (left, right, scope, target) => setVariable(left, scope.get(left) - right, scope, target),
  '*=': (left, right, scope, target) => setVariable(left, scope.get(left) * right, scope, target),
  '&=': (left, right, scope, target) => setVariable(left, scope.get(left) & right, scope, target),
  '|=': (left, right, scope, target) => setVariable(left, scope.get(left) | right, scope, target),
  '^=': (left, right, scope, target) => setVariable(left, scope.get(left) ^ right, scope, target),
  '<<': (left, right) => left << right,
  '>>': (left, right) => left >> right,
  or: (left, right) => left || right,
  eq: (left, right) => left === right,
  ne: (left, right) => left !== right,
  lt: (left, right) => left < right,
  le: (left, right) => left <= right,
  gt: (left, right) => left > right,
  ge: (left, right) => left >= right,
  and: (left, right) => left && right,
  xor: (left, right) => !!(left ^ right)
};
visitors.UnaryExpression = (node, scope, walk) =>
  unaryOperation[node.operator](walk(node.argument, node, scope));
const unaryOperation = {
  '!': value => !value,
  '+': value => value,
  '-': value => -value,
  '~': value => ~value,
  not: value => !value
};
visitors.MemberExpression = (node, scope, walk) => {
  const object = walk(node.object, node, scope);
  if (!object) throw new RuntimeError('object is undefined', node.object)
  if (!(typeof object === 'object' && object && !Array.isArray(object))) {
    throw new RuntimeError(`dereferenced source (${typeof array}) is not an object or assoc`, node.object)
  }
  let { property } = node;
  const { type } = property;
  property = type === 'Identifier' || type === 'LegacyAlias'
    ? property.value
    : walk(node.property, node, scope);
  return object[property]
};
visitors.SliceExpression = (node, scope, walk) => {
  const array = walk(node.object, node, scope);
  if (!array) throw new RuntimeError('slice source is undefined', node.object)
  if (!(typeof array === 'string' || Array.isArray(array))) {
    throw new RuntimeError(`slice source (${typeof array}) is not a list or string`, node.object)
  }
  const start = (node.start && walk(node.start, node, scope)) || 0;
  const end = (node.end && walk(node.end, node, scope)) || array.length;
  return array.slice(start, end)
};
visitors.IndexExpression = (node, scope, walk) => {
  const array = walk(node.object, node, scope);
  if (!array) throw new RuntimeError('index source is undefined', node.object)
  if (!(typeof array === 'string' || Array.isArray(array))) {
    throw new RuntimeError(`index source (${typeof array}) is not a list or string`, node.object)
  }
  return array[walk(node.index, node, scope)]
};
visitors.CallExpression = (node, scope, walk) => {
  const routine = walk(node.callee, node, scope);
  if (!routine) {
    const name = getName(node.callee);
    if (name) throw new RuntimeError(`function "${name}" is undefined`, node.callee)
    throw new RuntimeError('undefined is not a function', node.callee)
  }
  const values = node.arguments.map(arg => walk(arg, node, scope));
  if (typeof routine === 'function') return routine(...values)
  if (!routine.type) {
    const name = getName(node.callee);
    if (name) throw new RuntimeError(`"${name}" is not a function`, node.callee)
    throw new RuntimeError(`${typeof routine} is not a function`, node.callee)
  }
  const args = routine.params.reduce((args, param, index) => {
    const value = values[index];
    args[param.id.value] = value === undefined
      ? param.init
          ? walk(param.init, node, scope)
          : getDefaultValue(param.parameterType)
      : value;
    return args
  }, {});
  scope = new Scope(scope, args);
  scope.result = undefined;
  if (node.routines) for (const routine of node.routines) walk(routine, node, scope);
  for (const variable of node.variables) walk(variable, node, scope);
  for (const statement of node.statements) walk(statement, node, scope);
  return scope.result
};
visitors.ThisExpression = (node, scope) => scope;
visitors.SuperExpression = (node, { upper }) => upper;
visitors.AssocExpression = visitors.ObjectExpression = (node, scope, walk) =>
  node.properties.reduce((object, { key, value }) => {
    object[walk(key, node, scope)] = walk(value, node, scope);
    return object
  }, {});
visitors.ListExpression = (node, scope, walk) =>
  node.elements.reduce((result, element) => {
    const item = walk(element, node, scope);
    if (element.type === 'AtExpression') push.apply(result, item);
    else result.push(item);
    return result
  }, []);
visitors.ListComprehension = (node, scope, walk) => {
  const name = node.left.value;
  const values = walk(node.right, node, scope);
  return values.reduce((result, value) => {
    scope.set(name, value);
    if (!node.test || walk(node.test, node, scope)) {
      result.push(walk(node.expression, node, scope));
    }
    return result
  }, [])
};
visitors.AtExpression = visitors.ParenthesisExpression =
  (node, scope, walk) => walk(node.expression, node, scope);
visitors.XlateExpression = (node, scope) => {
  let { raw } = node;
  if (!raw) raw = prepareXlateExpression(node);
  scope.get(raw);
};
visitors.Identifier = ({ value }, scope) => scope.get(value);
visitors.LegacyAlias = (node, scope) => {
  let { raw } = node;
  if (!raw) raw = prepareLegacyAlias(node);
  scope.get(raw);
};
visitors.Literal = ({ value }) => value;
function prepareScript (node) {
  const routines = node.routines = [];
  const variables = node.variables = [];
  const statements = node.statements = [];
  for (const part of node.body) {
    const { type } = part;
    const target = type === 'VariableDeclaration'
      ? variables
      : type === 'FunctionDeclaration' || type === 'ScriptDeclaration'
        ? routines
        : statements;
    target.push(part);
  }
}
function prepareFunction (node) {
  const variables = node.variables = [];
  const statements = node.statements = [];
  for (const part of node.body) {
    const { type } = part;
    const target = type === 'VariableDeclaration' ? variables : statements;
    target.push(part);
  }
}
function prepareObjectName (node) {
  const raw = node.name
    .map(({ value }) => typeof value === 'number' ? `&${value.toString(16)}` : value)
    .join('::');
  node.raw = raw;
  return raw
}
function prepareLegacyAlias (node) {
  const raw = `&${node.value.toString(16)}`;
  node.raw = raw;
  return raw
}
function prepareXlateExpression (node) {
  const raw = `${node.ospace.value}.${node.string.value}`;
  node.raw = raw;
  return raw
}

const ascending = 1;
const descending = 2;
function copy (assoc) {
  checkAssoc(assoc, 1);
  return { ...assoc }
}
function createassoc () {
  return {}
}
function createfrompairs (...pairs) {
  const result = {};
  for (let i = 1, l = pairs.length; i < l; i += 2) result[pairs[i]] = pairs[i + 1];
  return result
}
function remove (assoc, key) {
  checkAssoc(assoc, 1);
  delete assoc[key];
}
function iskey (assoc, key) {
  checkAssoc(assoc, 1);
  return key in assoc
}
function items (assoc) {
  checkAssoc(assoc, 1);
  return Object.values(assoc)
}
function keys (assoc) {
  checkAssoc(assoc, 1);
  return Object.keys(assoc)
}
function merge (left, right) {
  checkAssoc(left, 1);
  checkAssoc(right, 2);
  return { ...left, ...right }
}
function undefinedvalue (assoc) {
  checkAssoc(assoc, 1);
}

var assoc = /*#__PURE__*/Object.freeze({
  __proto__: null,
  ascending: ascending,
  descending: descending,
  copy: copy,
  createassoc: createassoc,
  createfrompairs: createfrompairs,
  remove: remove,
  'delete': remove,
  iskey: iskey,
  items: items,
  keys: keys,
  merge: merge,
  undefinedvalue: undefinedvalue
});

const ascending$1 = 1;
const descending$1 = 2;
function allocate (size) {
  checkType(size, 'number', 1);
  return new Array(size)
}
function create (...values) {
  return values
}
function max (list) {
  checkList(list, 1);
  let max = list[0];
  for (let i = 1, l = list.length; i < l; ++i) {
    const value = list[i];
    if (value > max) max = value;
  }
  return max
}
function min (list) {
  checkList(list, 1);
  let min = list[0];
  for (let i = 1, l = list.length; i < l; ++i) {
    const value = list[i];
    if (value < min) min = value;
  }
  return min
}
function setadd (list, item) {
  checkList(list, 1);
  if (list.some(value => value === item)) return list
  return list.concat(item)
}
function setremove (list, item) {
  checkList(list, 1);
  const index = list.findIndex(value => value === item);
  if (index < 0) return list
  if (index === 0) return list.slice(1)
  const start = list.slice(0, index);
  return index === list.length - 1 ? start : start.concat(list.slice(index + 1))
}
function setunion (left, right) {
  checkList(left, 1);
  checkList(right, 2);
  const result = left.slice();
  for (const item of right) if (left.every(value => value !== item)) result.push(item);
  return result
}
const compares = [
  undefined,
  function ascending (left, right) {
    return left < right ? -1 : left > right ? 1 : 0
  },
  function descending (left, right) {
    return left < right ? 1 : left > right ? -1 : 0
  }
];
function sort (list, order) {
  checkList(list, 1);
  const compare = compares[order];
  if (!compare) throw new TypeError(`ASCENDING or DESCENDING expected, but ${typeof order}" found in the second argument`)
  return list.sort(compare)
}

var list = /*#__PURE__*/Object.freeze({
  __proto__: null,
  ascending: ascending$1,
  descending: descending$1,
  allocate: allocate,
  create: create,
  max: max,
  min: min,
  setadd: setadd,
  setremove: setremove,
  setunion: setunion,
  sort: sort
});

const cr = '\r';
const crlf = '\r\n';
const lf = '\n';
const tab = '\t';
function ascii (value) {
  checkType2(value, 'string', 'number', 1);
  return typeof value === 'number' ? String.fromCharCode(value) : value.charCodeAt(0)
}
function bytelength (text) {
  checkType(text, 'string', 1);
  return typeof Buffer !== 'undefined'
    ? Buffer.from(text).length
    : new TextEncoder().encode(text).length
}
function capitalize (text) {
  checkType(text, 'string', 1);
  return text.replace(/(\w+)/g, match =>
    match.charAt(0).toUpperCase() + match.substr(1).toLowerCase())
}
function catenate (values, delimiter) {
  checkList(values, 1);
  checkType(delimiter, 'string', 2);
  const suffix = values.length ? delimiter : '';
  return `${values.join(delimiter)}${suffix}`
}
function chr (hay, needle) {
  checkType(hay, 'string', 1);
  checkType(needle, 'string', 2);
  const index = hay.indexOf(needle.charAt(0));
  return index < 0 ? undefined : index + 1
}
function cmp (left, right, length) {
  checkType(left, 'string', 1);
  checkType(right, 'string', 2);
  checkTypeOptional(length, 'number', 3);
  if (length !== undefined) {
    left = left.substr(0, length);
    right = right.substr(0, length);
  }
  return left < right ? -1 : left > right ? 1 : 0
}
function cmpbe (left, right, length) {
  checkType(left, 'string', 1);
  checkType(right, 'string', 2);
  checkTypeOptional(length, 'number', 3);
  return cmp(left.trimEnd(), right.trimEnd(), length)
}
function cmpi (left, right, length) {
  checkType(left, 'string', 1);
  checkType(right, 'string', 2);
  checkTypeOptional(length, 'number', 3);
  return cmp(left.toLowerCase(), right.toLowerCase(), length)
}
function cmpibe (left, right, length) {
  checkType(left, 'string', 1);
  checkType(right, 'string', 2);
  checkTypeOptional(length, 'number', 3);
  return cmp(left.trimEnd().toLowerCase(), right.trimEnd().toLowerCase(), length)
}
function collapse (text) {
  checkType(text, 'string', 1);
  return text.replace(/[ \t]/g, '')
}
function compress (text) {
  checkType(text, 'string', 1);
  return text.replace(/[ \t]+/g, ' ')
}
function cspn (text, chars) {
  checkType(text, 'string', 1);
  checkType(chars, 'string', 2);
  const regexp = new RegExp(`^[^${chars.split('').map(escapeRegExp).join('|')}]+`);
  const match = regexp.exec(text);
  return match ? match[0].length : 0
}
function cstring (text, quote = '\'') {
  return text.replace(new RegExp(`[\r\n\t\\${quote}]`, 'g'), '\\$1')
}
function elements (text, delimiter) {
  checkType(text, 'string', 1);
  checkType(delimiter, 'string', 2);
  return text.split(delimiter.charAt(0))
}
function filetostring (name) {
  checkType(name, 'string', 1);
  return fs.readFileSync(name, 'utf8')
}
function format (format, ...args) {
  return format.replace(/%(\d)/g, (match, index) => '' + args[index - 1])
}
function frombase64 (text) {
  checkType(text, 'string', 1);
  return typeof Buffer !== 'undefined'
    ? Buffer.from(text, 'base64').toString('ascii')
    : atob(text)
}
function hyphenate (text, length) {
  checkType(text, 'string', 1);
  checkType(text, 'number', 2);
  const parts = [];
  do {
    parts.push(text.substr(0, length));
    text = text.substr(length);
  } while (text.length)
  return parts
}
function join (values, delimiter) {
  checkList(values, 1);
  checkType(delimiter, 'string', 2);
  return values.join(delimiter)
}
function locate$1 (hay, needle) {
  checkType(hay, 'string', 1);
  checkType(needle, 'string', 2);
  const index = hay.indexOf(needle);
  return index < 0 ? undefined : index + 1
}
function locatei (hay, needle) {
  checkType(hay, 'string', 1);
  checkType(needle, 'string', 2);
  const index = hay.toLowerCase().indexOf(needle.toLowerCase());
  return index < 0 ? undefined : index + 1
}
function lower (text) {
  checkType(text, 'string', 1);
  return text.toLowerCase()
}
function quote (text, quote = '\'') {
  checkType(text, 'string', 1);
  checkQuote(quote, 2);
  return `${quote}${text.replace(new RegExp(quote, 'g'), `\\${quote}`)}${quote}`
}
function rchr (hay, needle) {
  checkType(hay, 'string', 1);
  checkType(needle, 'string', 2);
  const index = hay.lastIndexOf(needle.charAt(0));
  return index < 0 ? undefined : index + 1
}
function replace (text, find, replace) {
  checkType(text, 'string', 1);
  checkType(find, 'string', 2);
  checkType(replace, 'string', 3);
  return text.replace(find, replace)
}
function replaceall (text, find, replace) {
  checkType(text, 'string', 1);
  checkType(find, 'string', 2);
  checkType(replace, 'string', 3);
  const regexp = new RegExp(`${find.split('').map(escapeRegExp).join('')}`, 'g');
  return text.replace(regexp, replace)
}
function set (length, text) {
  checkType2(text, 'string', 'number', 1);
  checkType(text, 'string', 2);
  if (typeof length === 'string') length = length.length;
  else if (length < 0) length = 0;
  return text.charAt(0).repeat(length)
}
function spn (text, chars) {
  checkType(text, 'string', 1);
  checkType(chars, 'string', 2);
  const regexp = new RegExp(`^[${chars.split('').map(escapeRegExp).join('')}]+`);
  const match = regexp.exec(text);
  return match ? match[0].length : 0
}
function string (value, raw) {
  checkTypeOptional(raw, 'boolean', 2);
  return !raw && value instanceof Error ? value.message : valuetostring(value)
}
function stringtointegher (text) {
  return parseInt(text)
}
function stringtoreal (text) {
  checkType(text, 'string', 1);
  return parseFloat(text)
}
function stringtovalue (text) {
  checkType(text, 'string', 1);
  const number = parseInt(text);
  if (isNaN(number)) return number
  const normalized = text.toLowerCase();
  if (normalized === 'true') return true
  if (normalized === 'false') return false
  if (normalized === 'undefined') return undefined
  return text
}
function strip (text, chars) {
  checkType(text, 'string', 1);
  checkType(chars, 'string', 2);
  const regexp = new RegExp(`[${chars.split('').map(escapeRegExp).join('')}]`, 'g');
  return text.replace(regexp, '')
}
function tobase64 (text) {
  checkType(text, 'string', 1);
  return typeof Buffer !== 'undefined'
    ? Buffer.from(text).toString('base64')
    : btoa(text)
}
function upper (text) {
  checkType(text, 'string', 1);
  return text.toUpperCase()
}
function valuetostring (value) {
  return String(value)
}
function escapeRegExp (char) {
  return char === '*' || char === '+' || char === '.' || char === '?' ||
      char === '[' || char === ']' || char === '(' || char === ')' ||
      char === '{' || char === '}' || char === '^' || char === '$'
    ? `\\${char}`
    : char
}

var str = /*#__PURE__*/Object.freeze({
  __proto__: null,
  cr: cr,
  crlf: crlf,
  lf: lf,
  tab: tab,
  ascii: ascii,
  bytelength: bytelength,
  capitalize: capitalize,
  catenate: catenate,
  chr: chr,
  cmp: cmp,
  cmpbe: cmpbe,
  cmpi: cmpi,
  cmpibe: cmpibe,
  collapse: collapse,
  compress: compress,
  cspn: cspn,
  cstring: cstring,
  elements: elements,
  filetostring: filetostring,
  format: format,
  frombase64: frombase64,
  hyphenate: hyphenate,
  join: join,
  locate: locate$1,
  locatei: locatei,
  lower: lower,
  quote: quote,
  rchr: rchr,
  replace: replace,
  replaceall: replaceall,
  set: set,
  spn: spn,
  string: string,
  stringtointegher: stringtointegher,
  stringtoreal: stringtoreal,
  stringtovalue: stringtovalue,
  strip: strip,
  tobase64: tobase64,
  upper: upper,
  valuetostring: valuetostring
});

const crlf$1 = '\r\n';
const js_encodeuri = 1;
const js_encodeuricomponent = 2;
const js_escape = 3;
function decodeforurl (text) {
  checkType(text, 'string', 1);
  return decodeURI(text)
}
function encodeforurl (text) {
  checkType(text, 'string', 1);
  return encodeURI(text)
}
function escape (text) {
  checkType(text, 'string', 1);
  return encodeURIComponent(text)
}
function escapeforjs (text, mode) {
  checkType(text, 'string', 1);
  if (mode === js_escape) return escape(text)
  if (mode === js_encodeuri) return encodeURI(text)
  if (mode === js_encodeuricomponent) return encodeURIComponent(text)
  throw new TypeError(`JS_ESCAPE, JS_ENCODEURI or JS_ENCODEURICOMPONENT expected, but ${typeof mode}" found in the second argument`)
}
function escapehtml (text) {
  checkType(text, 'string', 1);
  return text
    .replace(/&/g, '&amp;')
    .replace(/</g, '&lt;')
    .replace(/</g, '&gt;')
}
function escapejson (text, trim = true) {
  checkType(text, 'string', 1);
  checkTypeOptional(text, 'boolean', 2);
  if (trim) text = text.trim();
  const string = JSON.stringify(text);
  return string.substr(1, string.length - 2)
}
function escapexml (text) {
  checkType(text, 'string', 1);
  return text
    .replace(/&/g, '&amp;')
    .replace(/</g, '&lt;')
    .replace(/</g, '&gt;')
    .replace(/'/g, '&apos;')
}
function tojson (text, trim = true) {
  checkType(text, 'string', 1);
  checkTypeOptional(text, 'boolean', 2);
  return JSON.parse(trimStrings(text))
}
function unescape (text) {
  checkType(text, 'string', 1);
  return decodeURIComponent(text)
}
function unescapejson (text) {
  checkType(text, 'string', 1);
  return JSON.parse(`"${text}}"`)
}
function trimStrings (value) {
  if (typeof value === 'string') return value.trim()
  if (Array.isArray(value)) {
    for (let i = 0, l = value.length; i < l; ++i) value[i] = trimStrings(value[i]);
  }
  if (value && typeof value === 'object') {
    for (const key in value) value[key] = trimStrings(value[key]);
  }
  return value
}

var web = /*#__PURE__*/Object.freeze({
  __proto__: null,
  crlf: crlf$1,
  js_encodeuri: js_encodeuri,
  js_encodeuricomponent: js_encodeuricomponent,
  js_escape: js_escape,
  decodeforurl: decodeforurl,
  encodeforurl: encodeforurl,
  escape: escape,
  escapeforjs: escapeforjs,
  escapehtml: escapehtml,
  escapejson: escapejson,
  escapexml: escapexml,
  tojson: tojson,
  unescape: unescape,
  unescapejson: unescapejson
});

const { log, debug, error, info, warn } = console;
let stamp;
function pad2 (n) { return n < 10 ? `0${n}` : n }
function pad3 (n) { return n < 10 ? `00${n}` : n < 100 ? `0${n}` : n }
function formatStamp () {
  const date = new Date();
  return pad2((date.getMonth() + 1)) + '/' + pad2(date.getDate()) + '/' +
    (date.getYear() + 1900) + ' ' + pad2(date.getHours()) + ':' +
    pad2(date.getMinutes()) + ':' + pad2(date.gitSeconds()) + '.' +
    pad3(date.getMilliseconds())
}
function formatMessage (...args) {
  const message = args.join('');
  return stamp ? `${formatStamp()} ${message}` : message
}
function echo (...args) {
  log(formatMessage(...args));
}
function echodebug (...args) {
  debug(formatMessage(...args));
}
function echoerror (...args) {
  error(formatMessage(...args));
}
function echoinfo (...args) {
  info(formatMessage(...args));
}
function echostamp (enable) {
  checkType(enable, 'boolean', 1);
  stamp = enable;
}
function echowarn (...args) {
  warn(formatMessage(...args));
}
function isdefined (value) {
  return value !== undefined
}
function iserror (value) {
  return value instanceof Error
}
function isfeature (object, property) {
  return property in object
}
function isinvokable (value) {
  return typeof value === 'function' || !!value.type
}
function isnoterror (value) {
  return !(value instanceof Error)
}
function isundefined (value) {
  return value === undefined
}
function length (value) {
  return value && value.length
}

var library = /*#__PURE__*/Object.freeze({
  __proto__: null,
  assoc: assoc,
  list: list,
  str: str,
  web: web,
  echo: echo,
  echodebug: echodebug,
  echoerror: echoerror,
  echoinfo: echoinfo,
  echostamp: echostamp,
  echowarn: echowarn,
  isdefined: isdefined,
  iserror: iserror,
  isfeature: isfeature,
  isinvokable: isinvokable,
  isnoterror: isnoterror,
  isundefined: isundefined,
  length: length
});

function interpret (ast, options = {}) {
  setOptions(options);
  const scope = new Scope(null, Object.assign({}, library, options.globals));
  return oscriptAstWalker.recursive(ast, null, visitors, scope)
}

exports.interpret = interpret;
//# sourceMappingURL=index.js.map
